--------------------------------------------------------------------------------
-- CERN (BE-CO-HT)
-- Top level entity of CONV-TTL-BLO V1
-- http://www.ohwr.org/projects/conv-trig-blo
--------------------------------------------------------------------------------
--
-- unit name: conv_ttl_blo_v2.vhd
--
-- author: Theodor-Adrian Stana (t.stana@cern.ch)
--
-- version: 1.0
--
-- description: Top entity of CONV-TTL-BLO V1
--
-- dependencies:
--
--------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
--------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--------------------------------------------------------------------------------
library IEEE;
library unisim;

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use UNISIM.VCOMPONENTS.ALL;

use work.bicolor_led_ctrl_pkg.all;
use work.wishbone_pkg.all;

entity conv_ttl_blo_v2 is
  generic
  (
    g_nr_ttl_chan : natural := 6;
    g_nr_inv_chan : natural := 4
  );
  port 
  (
    -- Clock inputs
    clk20_vcxo_i           : in  std_logic;
    fpga_clk_p_i           : in  std_logic; --Using the 125MHz clock
    fpga_clk_n_i           : in  std_logic;
    
    -- LEDs
    led_ctrl0_o            : out std_logic;
    led_ctrl0_oen_o        : out std_logic;
    led_ctrl1_o            : out std_logic;
    led_ctrl1_oen_o        : out std_logic;
    led_multicast_2_0_o    : out std_logic;
    led_multicast_3_1_o    : out std_logic;
    led_wr_gmt_ttl_ttln_o  : out std_logic;
    led_wr_link_syserror_o : out std_logic;
    led_wr_ok_syspw_o      : out std_logic;
    led_wr_ownaddr_i2c_o   : out std_logic;
    
    -- I/Os for pulses
    pulse_front_led_n_o    : out std_logic_vector(g_nr_ttl_chan downto 1);
    pulse_rear_led_n_o     : out std_logic_vector(g_nr_ttl_chan downto 1);
--    fpga_input_ttl_n_i     : in  std_logic_vector(g_nr_ttl_chan downto 1);
--    fpga_out_ttl_o         : out std_logic_vector(g_nr_ttl_chan downto 1);
--    fpga_blo_in_i          : in  std_logic_vector(g_nr_ttl_chan downto 1);
--    fpga_trig_blo_o        : out std_logic_vector(g_nr_ttl_chan downto 1);
--    inv_in_n_i             : in  std_logic_vector(g_nr_inv_chan downto 1);
--    inv_out_o              : out std_logic_vector(g_nr_inv_chan downto 1);
    
    -- Lines for the i2c_slave
    scl_i                  : in  std_logic;
    scl_o                  : out std_logic;
    scl_oe_o               : out std_logic;
    sda_i                  : in  std_logic;
    sda_o                  : out std_logic;
    sda_oe_o               : out std_logic;
    fpga_ga_i              : in  std_logic_vector(4 downto 0);
    fpga_gap_i             : in  std_logic;
--    
--    -- Pins of the SPI interface to write into the Flash memory
--    fpga_prom_cclk_o       : out std_logic;
--    fpga_prom_cso_b_n_o    : out std_logic;
--    fpga_prom_miso_i       : in  std_logic;
--    fpga_prom_mosi_o       : out std_logic;
--      
    fpga_oe_o              : out std_logic;
    fpga_blo_oe_o          : out std_logic;
    fpga_trig_ttl_oe_o     : out std_logic;
    fpga_inv_oe_o          : out std_logic;
    
    -- PLL DAC control pins
    fpga_plldac1_sclk_o    : out std_logic;
    fpga_plldac1_sync_n_o  : out std_logic;
    fpga_plldac1_din_o     : out std_logic;

    -- VCXO DAC control pins
    fpga_plldac2_sclk_o    : out std_logic;
    fpga_plldac2_sync_n_o  : out std_logic;
    fpga_plldac2_din_o     : out std_logic;

    -- SFP I2C lines
    fpga_sfp_mod_def1_b    : inout std_logic;
    fpga_sfp_mod_def2_b    : inout std_logic;

    -- TTL/INV_TTL_N
    ttl_switch_n_i         : in  std_logic;
    
--    extra_switch_n_i         : in  std_logic_vector(7 downto 1);
    
    -- It allows power sequencing of the
    -- 24V rail after a security given delay
    mr_n_o                 : out std_logic;

    -- RTM identifiers, should match with the expected values
    -- TODO: add matching
    fpga_rtmm_n_i          : in std_logic_vector(2 downto 0);
    fpga_rtmp_n_i          : in std_logic_vector(2 downto 0);

    -- DS18B20U+ thermometer bidirectional port
    thermometer_b          : inout std_logic
  );
end conv_ttl_blo_v2;


architecture behav of conv_ttl_blo_v2 is

  --============================================================================
  -- Type declarations
  --============================================================================
  type t_pts_state is
    (
      ST_IDLE,
      ST_PLLTEST,
      ST_LEDTEST_1,
      ST_LEDTEST_2
    );
  --============================================================================
  -- Constant declarations
  --============================================================================
  -- Number of Wishbone masters and slaves, for wb_crossbar
  constant c_nr_masters : NATURAL := 1;
  constant c_nr_slaves  : NATURAL := 7;

  -----------------------------------------
  -- Memory map
  -- * all registers are word-addressable
  -- * all registers are word aligned
  -----------------------------------------
  --    PTS           [0000-0010]
  --    ONEWIRE       [0010-0020]
  --    DAC_SPI_125   [0020-0080]
  --    CLK_INFO_125  [0080-0100]
  --    DAC_SPI_20    [0100-0160]
  --    CLK_INFO_20   [0160-0180]
  --    SFP_I2C       [0180-0200]
  -----------------------------------------
  -- slave order definitions
  constant c_slv_pts_regs      : natural := 0;
  constant c_slv_onewire_mst   : natural := 1;
  constant c_slv_dac_spi_125   : natural := 2;
  constant c_slv_clk_info_125  : natural := 3;
  constant c_slv_dac_spi_20    : natural := 4;
  constant c_slv_clk_info_20   : natural := 5;
  constant c_slv_sfp_i2c       : natural := 6;

  -- base address definitions
  constant c_addr_pts_regs     : t_wishbone_address := x"00000000";
  constant c_addr_onewire_mst  : t_wishbone_address := x"00000010";
  constant c_addr_dac_spi_125  : t_wishbone_address := x"00000020";
  constant c_addr_clk_info_125 : t_wishbone_address := x"00000080";
  constant c_addr_dac_spi_20   : t_wishbone_address := x"00000100";
  constant c_addr_clk_info_20  : t_wishbone_address := x"00000160";
  constant c_addr_sfp_i2c      : t_wishbone_address := x"00000180";

  -- address mask definitions
  constant c_mask_pts_regs     : t_wishbone_address := x"FFFFFFF0"; 
  constant c_mask_onewire_mst  : t_wishbone_address := x"FFFFFFF0";
  constant c_mask_dac_spi_125  : t_wishbone_address := x"FFFFFFE0";
  constant c_mask_clk_info_125 : t_wishbone_address := x"FFFFFFC0";
  constant c_mask_dac_spi_20   : t_wishbone_address := x"FFFFFFE0";
  constant c_mask_clk_info_20  : t_wishbone_address := x"FFFFFFE0";
  constant c_mask_sfp_i2c      : t_wishbone_address := x"FFFFFFE0";
  
  -- addresses constant for Wishbone crossbar
  constant c_addresses  : t_wishbone_address_array(c_nr_slaves-1 downto 0)
                        := (
                             c_slv_pts_regs     => c_addr_pts_regs,
                             c_slv_onewire_mst  => c_addr_onewire_mst,
                             c_slv_dac_spi_125  => c_addr_dac_spi_125,
                             c_slv_clk_info_125 => c_addr_clk_info_125,
                             c_slv_dac_spi_20   => c_addr_dac_spi_20,
                             c_slv_clk_info_20  => c_addr_clk_info_20,
                             c_slv_sfp_i2c      => c_addr_sfp_i2c
                           );

  -- masks constant for Wishbone crossbar
  constant c_masks      : t_wishbone_address_array(c_nr_slaves-1 downto 0)
                        := (
                             c_slv_pts_regs     => c_mask_pts_regs,
                             c_slv_onewire_mst  => c_mask_onewire_mst,
                             c_slv_dac_spi_125  => c_mask_dac_spi_125,
                             c_slv_clk_info_125 => c_mask_clk_info_125,
                             c_slv_dac_spi_20   => c_mask_dac_spi_20,
                             c_slv_clk_info_20  => c_mask_clk_info_20,
                             c_slv_sfp_i2c      => c_mask_sfp_i2c
                           );

  --============================================================================
  -- Component declarations
  --============================================================================
  -- Reset generator component
  -- (use: global reset generation, output reset generation)
  component reset_gen is
    generic
    (
      -- Reset time in number of clk_i cycles
      g_reset_time : positive := 5_000_000
    );
    port
    (
      clk_i   : in  std_logic;
      rst_i   : in  std_logic;
      rst_n_o : out std_logic
    );
  end component reset_gen;

  -- RTM detector component
  -- (use: detect the presence of an RTM/P module)
  component rtm_detector is
    port 
    (
      rtmm_i    : in std_logic_vector(2 downto 0);
      rtmp_i    : in std_logic_vector(2 downto 0);
      rtmm_ok_o : out std_logic;
      rtmp_ok_o : out std_logic
    );
  end component rtm_detector;

  -- I2C bridge
  -- (use: convert I2C transfers into WB transfers on memmapped registers)  
  component vme64x_i2c is
    port
    (
      -- Clock, reset
      clk_i      : in  std_logic;
      rst_n_i    : in  std_logic;

      -- I2C lines
      sda_en_o   : out std_logic;
      sda_i      : in  std_logic;
      sda_o      : out std_logic;
      scl_en_o   : out std_logic;
      scl_i      : in  std_logic;
      scl_o      : out std_logic;

      -- I2C address and status
      i2c_addr_i : in  std_logic_vector(6 downto 0);
      i2c_done_o : out std_logic;

      -- Wishbone master signals
      wbm_stb_o  : out std_logic;
      wbm_cyc_o  : out std_logic;
      wbm_sel_o  : out std_logic_vector(3 downto 0);
      wbm_we_o   : out std_logic;
      wbm_dat_i  : in  std_logic_vector(31 downto 0);
      wbm_dat_o  : out std_logic_vector(31 downto 0);
      wbm_adr_o  : out std_logic_vector(31 downto 0);
      wbm_ack_i  : in  std_logic;
      wbm_rty_i  : in  std_logic;
      wbm_err_i  : in  std_logic
    );
  end component vme64x_i2c;
  
  -- Regs to test I2C operation
  component pts_regs is
    port (
      rst_n_i                                  : in     std_logic;
      clk_sys_i                                : in     std_logic;
      wb_adr_i                                 : in     std_logic_vector(0 downto 0);
      wb_dat_i                                 : in     std_logic_vector(31 downto 0);
      wb_dat_o                                 : out    std_logic_vector(31 downto 0);
      wb_cyc_i                                 : in     std_logic;
      wb_sel_i                                 : in     std_logic_vector(3 downto 0);
      wb_stb_i                                 : in     std_logic;
      wb_we_i                                  : in     std_logic;
      wb_ack_o                                 : out    std_logic;
      wb_stall_o                               : out    std_logic;
      -- Port for std_logic_vector field: 'current test' in reg: 'control register'
      pts_ctrl_crrt_test_o                     : out    std_logic_vector(3 downto 0);
      -- Port for BIT field: 'reset' in reg: 'control register'
      pts_ctrl_rst_o                           : out    std_logic;
      -- Port for std_logic_vector field: 'bits' in reg: 'Board ID'
      pts_id_bits_o                            : out    std_logic_vector(31 downto 0)
    );
  end component pts_regs;
  
  component incr_counter is
    generic
    (
      width             : integer := 32 -- default size
    );
    port
    (
      -- INPUTS
      -- Signals from the clk_rst_manager
      clk_i             : in std_logic;
      rst_i             : in std_logic;

      -- Signals from any unit
      counter_top_i     : in std_logic_vector(width-1 downto 0); -- max value to be counted; when reached
                                                                 -- counter stays at it, until a reset
      counter_incr_en_i : in std_logic;   -- enables counting

     -- OUTPUTS
     -- Signals to any unit
      counter_o         : out std_logic_vector(width-1 downto 0);
      counter_is_full_o : out std_logic
    ); -- counter reahed counter_top_i value
  end component incr_counter;

  component clk_info_wb_slave is
    port
    (
      -- WISHBONE slave signals
      wb_clk_i          : in  std_logic;
      rst_i             : in  std_logic;
      wb_cyc_i          : in  std_logic;
      wb_stb_i          : in  std_logic;
      wb_addr_i         : in  std_logic_vector( 2 downto 0);
      wb_data_i         : in  std_logic_vector(31 downto 0);
      wb_we_i           : in  std_logic;
      wb_data_o         : out std_logic_vector(31 downto 0);
      wb_ack_o          : out std_logic;
       -- Clock info signals
      counter_is_full_i : in std_logic;
      counter_now_i     : in std_logic_vector (31 downto 0);
      counter_top_o     : out std_logic_vector (31 downto 0);
      counter_rst_o     : out std_logic;
      counter_en_o      : out std_logic;
      oe_clk_o          : out std_logic
    );
  end component clk_info_wb_slave;

  --============================================================================
  -- Signal declarations
  --============================================================================
  -- Clock signals
  signal clk125              : std_logic;

  -- Reset signals
  signal rst_n, rst          : std_logic;
  signal rst_fr_reg          : std_logic;

  -- RTM detection signals
  signal rtmm, rtmp          : std_logic_vector(2 downto 0);
  signal rtmm_ok, rtmp_ok    : std_logic;

  -- Signals for pulse generation triggers
  signal trig                : std_logic_vector(g_nr_ttl_chan downto 1);
  signal trig_inv            : std_logic_vector(g_nr_inv_chan downto 1);
  signal trig_ttl, trig_blo  : std_logic_vector(g_nr_ttl_chan downto 1);

  -- Temporary signal for blocking and TTL pulse outputs
  signal pulse_outputs       : std_logic_vector(g_nr_ttl_chan downto 1);
  
  -- Temporary signal for inverted-TTL pulse outputs
  signal inv_outputs         : std_logic_vector(g_nr_inv_chan downto 1);
  
  -- Pulse status LED signals
  signal front_led_en        : std_logic_vector(g_nr_ttl_chan downto 1);
  signal rear_led_en         : std_logic_vector(g_nr_ttl_chan downto 1);
  signal pulse_leds          : std_logic_vector(g_nr_ttl_chan downto 1);

  -- Output enable signals
  signal oe, ttl_oe          : std_logic;
  signal blo_oe, inv_oe      : std_logic;

  -- Signal for controlling the bicolor LED matrix
  signal bicolor_led_state   : std_logic_vector(23 downto 0);

  -- Wishbone crossbar signals
  signal xbar_slave_in       : t_wishbone_slave_in_array   (c_nr_masters - 1 downto 0);
  signal xbar_slave_out      : t_wishbone_slave_out_array  (c_nr_masters - 1 downto 0);
  signal xbar_master_in      : t_wishbone_master_in_array  (c_nr_slaves  - 1 downto 0);
  signal xbar_master_out     : t_wishbone_master_out_array (c_nr_slaves  - 1 downto 0);
  
  -- I2C bridge signals
  signal i2c_done            : std_logic;
  signal i2c_up              : std_logic;
  signal i2c_addr            : std_logic_vector(6 downto 0);

  -- LED signals
  signal led_seq             : unsigned( 4 downto 0);
  signal front_led_ledtest   : std_logic_vector(g_nr_ttl_chan downto 1);
  signal front_led_plltest   : std_logic_vector(g_nr_ttl_chan downto 1);

  -- one-wire master signals
  signal owr_pwren           : std_logic_vector(0 downto 0);
  signal owr_en              : std_logic_vector(0 downto 0);
  signal owr_in              : std_logic_vector(0 downto 0);

  -- PTS FSM signals
  signal pts_state           : t_pts_state;
  signal pts_crrt_test       : std_logic_vector(3 downto 0);
  signal cnt_halfsec         : unsigned(25 downto 0);

  -- PLL & DAC test signals
  signal cnt125              : std_logic_vector(31 downto 0);
  signal cnt125_top          : std_logic_vector(31 downto 0);
  signal cnt125_en           : std_logic;
  signal cnt125_rst          : std_logic;
  signal cnt125_full         : std_logic;
  signal cnt125_actual_rst   : std_logic;
  signal plldac1_sync_n      : std_logic_vector(7 downto 0);
  
  signal cnt20               : std_logic_vector(31 downto 0);
  signal cnt20_top           : std_logic_vector(31 downto 0);
  signal cnt20_en            : std_logic;
  signal cnt20_rst           : std_logic;
  signal cnt20_full          : std_logic;
  signal cnt20_actual_rst    : std_logic;
  signal cnt20_actual_rst_d0 : std_logic;
  signal plldac2_sync_n      : std_logic_vector(7 downto 0);

  -- SFP EEPROM signals
  signal i2c_scl_fr_sfp      : std_logic;
  signal i2c_scl_to_sfp      : std_logic;
  signal i2c_sfp_scl_en      : std_logic;

  signal i2c_sda_fr_sfp      : std_logic;
  signal i2c_sda_to_sfp      : std_logic;
  signal i2c_sfp_sda_en      : std_logic;

begin

  --============================================================================
  -- Generate 125 MHz global signal from differential lines
  --============================================================================
  cmp_125_diff_buf: IBUFGDS
    generic map 
    (
      DIFF_TERM    => TRUE,
      IBUF_LOW_PWR => TRUE
    )
    port map 
    (
      I  => fpga_clk_p_i,
      IB => fpga_clk_n_i,
      O  => clk125
    );

  --============================================================================
  -- Internal and external reset generation
  --============================================================================
  -- Configure reset generator for 96ms power-on reset
  cmp_reset_gen: reset_gen
    generic map
    (
      -- Reset time: 12 * 8ns * (10**6) = 96 ms
      g_reset_time => 12*(10**6)
    )
    port map
    (
      clk_i   => clk125,
      rst_i   => rst_fr_reg,
      rst_n_o => rst_n
    );
  
  rst    <= not rst_n;
  mr_n_o <= rst_n;
  
  --============================================================================
  -- I2C bridge logic
  --============================================================================
  i2c_addr <= "10" & fpga_ga_i;
  
  cmp_i2c_bridge: vme64x_i2c
    port map
    (
      -- Clock, reset
      clk_i      => clk125,
      rst_n_i    => rst_n,

      -- I2C lines
      sda_en_o   => sda_oe_o,
      sda_i      => sda_i,
      sda_o      => sda_o,
      scl_en_o   => scl_oe_o,
      scl_i      => scl_i,
      scl_o      => scl_o,

      -- I2C address and status
      i2c_addr_i => i2c_addr,
      i2c_done_o => i2c_done,

      -- Wishbone master signals
      wbm_stb_o  => xbar_slave_in(0).stb,
      wbm_cyc_o  => xbar_slave_in(0).cyc,
      wbm_sel_o  => xbar_slave_in(0).sel,
      wbm_we_o   => xbar_slave_in(0).we,
      wbm_dat_i  => xbar_slave_out(0).dat,
      wbm_dat_o  => xbar_slave_in(0).dat,
      wbm_adr_o  => xbar_slave_in(0).adr,
      wbm_ack_i  => xbar_slave_out(0).ack,
      wbm_rty_i  => xbar_slave_out(0).rty,
      wbm_err_i  => xbar_slave_out(0).err
    );
  
  -- Process to set the I2C_UP signal for display on the front panel
  -- of the front module. The I2C_UP signal is permanently set once an
  -- I2C transfer has successfully completed, as signaled by the RD_DONE
  -- and WR_DONE outputs of the I2C slave.
  p_i2c_up: process (clk125) is
  begin
    if rising_edge(clk125) then
      if (rst_n = '0') then
        i2c_up <= '0';
      elsif (i2c_done = '1') then
        i2c_up <= '1';
      end if;
    end if;
  end process p_i2c_up;

  --============================================================================
  -- Instantiation and connection the main Wishbone crossbar
  --============================================================================
  xbar_master_in(0).int <= '0';
  xbar_master_in(0).err <= '0';

  cmp_wb_crossbar: xwb_crossbar
  generic map
  (
    g_num_masters => c_nr_masters,
    g_num_slaves  => c_nr_slaves,
    g_registered  => false,
    g_address     => c_addresses,
    g_mask        => c_masks
  )
  port map
  (
    clk_sys_i => clk125,
    rst_n_i   => rst_n,
    slave_i   => xbar_slave_in,
    slave_o   => xbar_slave_out,
    master_i  => xbar_master_in,
    master_o  => xbar_master_out
  );

  --============================================================================
  -- PTS registers
  --============================================================================
  -- Regs to test I2C operation
  cmp_pts_regs: pts_regs
    port map
    (
      rst_n_i        => rst_n,
      clk_sys_i      => clk125,
      wb_adr_i       => xbar_master_out(c_slv_pts_regs).adr(2 downto 2),
      wb_dat_i       => xbar_master_out(c_slv_pts_regs).dat,
      wb_dat_o       => xbar_master_in(c_slv_pts_regs).dat,
      wb_cyc_i       => xbar_master_out(c_slv_pts_regs).cyc,
      wb_sel_i       => xbar_master_out(c_slv_pts_regs).sel,
      wb_stb_i       => xbar_master_out(c_slv_pts_regs).stb,
      wb_we_i        => xbar_master_out(c_slv_pts_regs).we,
      wb_ack_o       => xbar_master_in(c_slv_pts_regs).ack,
      wb_stall_o     => xbar_master_in(c_slv_pts_regs).stall,

      -- PTS control register
      pts_ctrl_crrt_test_o => pts_crrt_test,
      pts_ctrl_rst_o       => rst_fr_reg,

      -- PTS ID register
      pts_id_bits_o => open
    );

  --============================================================================
  -- FSM logic
  --============================================================================
  -- The test number is a four-bit fractional with one bit fractional part. It
  -- is set via the first four bits of the PTS control register and is used
  -- to set the state of the FSM.
  pts_state <= ST_PLLTEST   when (pts_crrt_test = "1000") else
               ST_LEDTEST_1 when (pts_crrt_test = "1110") else
               ST_LEDTEST_2 when (pts_crrt_test = "1111") else
               ST_IDLE;
  
  -- This process implements the main FSM of the PTS firmware, governing
  -- operation during various tests. The operations implemented are shown
  -- below:
  --
  --    test07          Increment LED sequence every time interval.
  p_fsm: process(clk125) is
  begin
    if rising_edge(clk125) then
      if (rst_n = '0') then
        -- pts_state   <= ST_IDLE;
        cnt_halfsec <= (others => '0');
        led_seq     <= (others => '0');
      else
        case pts_state is
          
          when ST_PLLTEST => 
            cnt_halfsec <= cnt_halfsec + 1;
            if (cnt_halfsec = 62499999) then
              cnt_halfsec <= (others => '0');
              led_seq     <= led_seq + 1;
              if (led_seq = 7) then
                led_seq   <= (others => '0');
              end if;
            end if;
            
          when ST_LEDTEST_1 =>
            cnt_halfsec <= cnt_halfsec + 1;
            if (cnt_halfsec = 62499999) then
              cnt_halfsec <= (others => '0');
              led_seq     <= led_seq + 1;
              if (led_seq = 7) then
                led_seq   <= (others => '0');
              end if;
            end if;

          when ST_LEDTEST_2 =>
            cnt_halfsec <= cnt_halfsec + 1;
            if (cnt_halfsec = 62499999) then
              cnt_halfsec <= (others => '0');
              led_seq     <= led_seq + 1;
              if (led_seq = 31) then
                led_seq   <= to_unsigned(8,5);
              end if;
            end if;

          when others =>
            null;

        end case;
      end if;
    end if;
  end process p_fsm;

  --============================================================================
  -- PLL test logic
  -- * test OSC3 -- 20MHz VCXO
  -- * test IC13 -- 125MHz PLL from TI
  -- * test IC17, IC18 -- Analog Devices AD5662 DACs used to control operation 
  --   of OSC1 & OSC3
  --============================================================================
  -----------------------------------------------------------------------------
  -- 125-MHz PLL test
  -----------------------------------------------------------------------------
  -- 125-MHz counter reset signal is OR between global reset and register reset
  cnt125_actual_rst <= cnt125_rst or rst;

  -- Instantiate increment counter for IC13
  cmp_pll125_incr_counter: incr_counter
    generic map
    (
      width => 32
    )
    port map
    (
      clk_i             => clk125,
      rst_i             => cnt125_actual_rst,
      counter_top_i     => cnt125_top,
      counter_incr_en_i => cnt125_en,
      counter_o         => cnt125,
      counter_is_full_o => cnt125_full
    );

  -- Instantiate clock info slave
  cmp_pll125_clk_info: clk_info_wb_slave
    port map
    (
      wb_clk_i          => clk125,
      rst_i             => rst,

      wb_cyc_i          => xbar_master_out(c_slv_clk_info_125).cyc,
      wb_stb_i          => xbar_master_out(c_slv_clk_info_125).stb,
      wb_addr_i         => xbar_master_out(c_slv_clk_info_125).adr(4 downto 2),
      wb_data_i         => xbar_master_out(c_slv_clk_info_125).dat,
      wb_we_i           => xbar_master_out(c_slv_clk_info_125).we,
      wb_data_o         => xbar_master_in(c_slv_clk_info_125).dat,
      wb_ack_o          => xbar_master_in(c_slv_clk_info_125).ack, 

      counter_is_full_i => cnt125_full,
      counter_now_i     => cnt125,
      counter_top_o     => cnt125_top,
      counter_rst_o     => cnt125_rst,
      counter_en_o      => cnt125_en,
      oe_clk_o          => open
    );

  -- Instantate a Wishbone SPI module to control IC18 DAC operation
  cmp_dac_125_spi: wb_spi
    generic map
    (
      g_interface_mode      => CLASSIC,
      g_address_granularity => WORD
    )
    port map
    (
      clk_sys_i  => clk125,
      rst_n_i    => rst_n,

      wb_adr_i   => xbar_master_out(c_slv_dac_spi_125).adr(6 downto 2),
      wb_dat_i   => xbar_master_out(c_slv_dac_spi_125).dat,
      wb_dat_o   => xbar_master_in(c_slv_dac_spi_125).dat,
      wb_sel_i   => xbar_master_out(c_slv_dac_spi_125).sel,
      wb_stb_i   => xbar_master_out(c_slv_dac_spi_125).stb,
      wb_cyc_i   => xbar_master_out(c_slv_dac_spi_125).cyc,
      wb_we_i    => xbar_master_out(c_slv_dac_spi_125).we,
      wb_ack_o   => xbar_master_in(c_slv_dac_spi_125).ack,
      wb_err_o   => xbar_master_in(c_slv_dac_spi_125).err,
      wb_int_o   => xbar_master_in(c_slv_dac_spi_125).int,
      wb_stall_o => xbar_master_in(c_slv_dac_spi_125).stall,

      pad_cs_o   => plldac1_sync_n,
      pad_sclk_o => fpga_plldac1_sclk_o,
      pad_mosi_o => fpga_plldac1_din_o,
      pad_miso_i => '0'
    );
  
  -- Finally, assign the SYNC_N output to the DAC
  fpga_plldac1_sync_n_o <= plldac1_sync_n(0);

  -----------------------------------------------------------------------------
  -- 20-MHz VCXO test
  -----------------------------------------------------------------------------
  -- First, synchronize the reset signal into the 20 MHz clock domain
  p_sync_rst: process (clk20_vcxo_i)
  begin
    if rising_edge(clk20_vcxo_i) then
      cnt20_actual_rst_d0 <= cnt20_rst;
      cnt20_actual_rst    <= cnt20_actual_rst_d0;
    end if;
  end process p_sync_rst;

  -- Then, instantiate increment counter for VCXO OSC1
  cmp_vcxo20_incr_counter: incr_counter
    generic map
    (
      width => 32
    )
    port map
    (
      clk_i             => clk20_vcxo_i,
      rst_i             => cnt20_actual_rst,
      counter_top_i     => cnt20_top,
      counter_incr_en_i => cnt20_en,
      counter_o         => cnt20,
      counter_is_full_o => cnt20_full
    );

  -- Instantiate clock info slave
  cmp_vcxo20_clk_info: clk_info_wb_slave
    port map
    (
      wb_clk_i          => clk125,
      rst_i             => rst,

      wb_cyc_i          => xbar_master_out(c_slv_clk_info_20).cyc,
      wb_stb_i          => xbar_master_out(c_slv_clk_info_20).stb,
      wb_addr_i         => xbar_master_out(c_slv_clk_info_20).adr(4 downto 2),
      wb_data_i         => xbar_master_out(c_slv_clk_info_20).dat,
      wb_we_i           => xbar_master_out(c_slv_clk_info_20).we,
      wb_data_o         => xbar_master_in(c_slv_clk_info_20).dat,
      wb_ack_o          => xbar_master_in(c_slv_clk_info_20).ack, 

      counter_is_full_i => cnt20_full,
      counter_now_i     => cnt20,
      counter_top_o     => cnt20_top,
      counter_rst_o     => cnt20_rst,
      counter_en_o      => cnt20_en,
      oe_clk_o          => open
    );

  -- Instantate a Wishbone SPI module to control IC18 DAC operation
  cmp_dac_20_spi: wb_spi
    generic map
    (
      g_interface_mode      => CLASSIC,
      g_address_granularity => WORD
    )
    port map
    (
      clk_sys_i  => clk125,
      rst_n_i    => rst_n,

      wb_adr_i   => xbar_master_out(c_slv_dac_spi_20).adr(6 downto 2),
      wb_dat_i   => xbar_master_out(c_slv_dac_spi_20).dat,
      wb_dat_o   => xbar_master_in(c_slv_dac_spi_20).dat,
      wb_sel_i   => xbar_master_out(c_slv_dac_spi_20).sel,
      wb_stb_i   => xbar_master_out(c_slv_dac_spi_20).stb,
      wb_cyc_i   => xbar_master_out(c_slv_dac_spi_20).cyc,
      wb_we_i    => xbar_master_out(c_slv_dac_spi_20).we,
      wb_ack_o   => xbar_master_in(c_slv_dac_spi_20).ack,
      wb_err_o   => xbar_master_in(c_slv_dac_spi_20).err,
      wb_int_o   => xbar_master_in(c_slv_dac_spi_20).int,
      wb_stall_o => xbar_master_in(c_slv_dac_spi_20).stall,

      pad_cs_o   => plldac2_sync_n,
      pad_sclk_o => fpga_plldac2_sclk_o,
      pad_mosi_o => fpga_plldac2_din_o,
      pad_miso_i => '0'
    );
  
  -- Finally, assign the SYNC_N output to the DAC
  fpga_plldac2_sync_n_o <= plldac2_sync_n(0);

  p_shift_front_leds: process(clk125) is
  begin
    if rising_edge(clk125) then
      if (rst_n = '0') then
        front_led_plltest <= "111110";
      else
        front_led_plltest <= front_led_plltest(g_nr_ttl_chan-1 downto 1) &
                             front_led_plltest(g_nr_ttl_chan);
      end if;
    end if;
  end process p_shift_front_leds;

  --============================================================================
  -- SFP EEPROM test logic
  -- * test bicolor LEDs and its driving circuit (IC1)
  -- * test front panel LED logic and driving circuit (IC5)
  --============================================================================
  -- First, instantiate an I2C master to handle SFP communication
  cmp_sfp_eeprom_i2c: wb_i2c_master
    generic map
    (
      g_interface_mode      => CLASSIC,
      g_address_granularity => WORD
    )
    port map
    (
      clk_sys_i    => clk125,
      rst_n_i      => rst_n,
      wb_adr_i     => xbar_master_out(c_slv_sfp_i2c).adr(6 downto 2),
      wb_dat_i     => xbar_master_out(c_slv_sfp_i2c).dat,
      wb_we_i      => xbar_master_out(c_slv_sfp_i2c).we,
      wb_stb_i     => xbar_master_out(c_slv_sfp_i2c).stb,
      wb_sel_i     => xbar_master_out(c_slv_sfp_i2c).sel,
      wb_cyc_i     => xbar_master_out(c_slv_sfp_i2c).cyc,
      wb_ack_o     => xbar_master_in(c_slv_sfp_i2c).ack,
      wb_int_o     => xbar_master_in(c_slv_sfp_i2c).int,
      wb_dat_o     => xbar_master_in(c_slv_sfp_i2c).dat,
      scl_pad_i    => i2c_scl_fr_sfp,
      scl_pad_o    => i2c_scl_to_sfp,
      scl_padoen_o => i2c_sfp_scl_en,
      sda_pad_i    => i2c_sda_fr_sfp,
      sda_pad_o    => i2c_sda_to_sfp,
      sda_padoen_o => i2c_sfp_sda_en
    );

  -- and assign the ports and tri-state buffers
  fpga_sfp_mod_def1_b <= i2c_scl_to_sfp when (i2c_sfp_scl_en = '0') else 'Z';
  i2c_scl_fr_sfp      <= fpga_sfp_mod_def1_b;
  fpga_sfp_mod_def2_b <= i2c_sda_to_sfp when (i2c_sfp_sda_en = '0') else 'Z';
  i2c_sda_fr_sfp      <= fpga_sfp_mod_def2_b;
    
  --============================================================================
  -- LED test logic
  -- * test bicolor LEDs and its driving circuit (IC1)
  -- * test front panel LED logic and driving circuit (IC5)
  --============================================================================
  -- The front LED outputs are multiplexed between the PLL test and the LED test.
  -- pulse_front_led_n_o <= front_led_plltest when (state = ST_TEST04) else
  --                        front_led_ledtest when (state = ST_TEST11) else
  --                        (others => '1');

  -- Sequence the front-panel LEDs based on the sequence counter in the FSM
  pulse_front_led_n_o <= "111110" when (led_seq =  1) else
                         "111101" when (led_seq =  2) else 
                         "111011" when (led_seq =  3) else 
                         "110111" when (led_seq =  4) else 
                         "101111" when (led_seq =  5) else 
                         "011111" when (led_seq =  6) else 
--                         "111110" when (led_seq =  7) else
--                         "111101" when (led_seq =  8) else 
--                         "111011" when (led_seq =  9) else 
--                         "110111" when (led_seq = 10) else 
--                         "101111" when (led_seq = 11) else 
--                         "011111" when (led_seq = 12) else 
--                         "111110" when (led_seq = 13) else
--                         "111101" when (led_seq = 14) else 
--                         "111011" when (led_seq = 15) else 
--                         "110111" when (led_seq = 16) else 
--                         "101111" when (led_seq = 17) else 
--                         "011111" when (led_seq = 18) else 
--                         "111110" when (led_seq = 19) else
--                         "111101" when (led_seq = 20) else 
--                         "111011" when (led_seq = 21) else 
--                         "110111" when (led_seq = 22) else 
--                         "101111" when (led_seq = 23) else 
--                         "011111" when (led_seq = 24) else 
                         "111111";

  pulse_rear_led_n_o <= "111110" when (led_seq =  1) else
                        "111101" when (led_seq =  2) else 
                        "111011" when (led_seq =  3) else 
                        "110111" when (led_seq =  4) else 
                        "101111" when (led_seq =  5) else 
                        "011111" when (led_seq =  6) else 
--                        "111110" when (led_seq =  7) else
--                        "111101" when (led_seq =  8) else 
--                        "111011" when (led_seq =  9) else 
--                        "110111" when (led_seq = 10) else 
--                        "101111" when (led_seq = 11) else 
--                        "011111" when (led_seq = 12) else 
--                        "111110" when (led_seq = 13) else
--                        "111101" when (led_seq = 14) else 
--                        "111011" when (led_seq = 15) else 
--                        "110111" when (led_seq = 16) else 
--                        "101111" when (led_seq = 17) else 
--                        "011111" when (led_seq = 18) else 
--                        "111110" when (led_seq = 19) else
--                        "111101" when (led_seq = 20) else 
--                        "111011" when (led_seq = 21) else 
--                        "110111" when (led_seq = 22) else 
--                        "101111" when (led_seq = 23) else 
--                        "011111" when (led_seq = 24) else 
                        "111111";

  -- Light each LED red and green in a sequence, based on the sequence counter
  -- in the FSM.
  --
  -- The colors are set via the LED state vector (two bits per LED) as follows:
  --    state(1..0)     color
  --        00           OFF
  --        01           RED
  --        10          GREEN
  bicolor_led_state <= "000000000001000000000000" when (led_seq =  8) else
                       "000000000100000000000000" when (led_seq =  9) else
                       "000000010000000000000000" when (led_seq = 10) else
                       "000001000000000000000000" when (led_seq = 11) else
                       "000000000000000001000000" when (led_seq = 12) else
                       "000000000000000000010000" when (led_seq = 13) else
                       "000000000000000000000100" when (led_seq = 14) else
                       "000000000000000000000001" when (led_seq = 15) else
                       "000000000000000100000000" when (led_seq = 16) else
                       "000000000000010000000000" when (led_seq = 17) else
                       "000100000000000000000000" when (led_seq = 18) else
                       "010000000000000000000000" when (led_seq = 19) else
                       "000000000010000000000000" when (led_seq = 20) else
                       "000000001000000000000000" when (led_seq = 21) else
                       "000000100000000000000000" when (led_seq = 22) else
                       "000010000000000000000000" when (led_seq = 23) else
                       "000000000000000010000000" when (led_seq = 24) else
                       "000000000000000000100000" when (led_seq = 25) else
                       "000000000000000000001000" when (led_seq = 26) else
                       "000000000000000000000010" when (led_seq = 27) else
                       "000000000000001000000000" when (led_seq = 28) else
                       "000000000000100000000000" when (led_seq = 29) else
                       "001000000000000000000000" when (led_seq = 30) else
                       "100000000000000000000000" when (led_seq = 31) else
                       "000000000000000000000000";

  -- Then, we instantiate the LED controller and control it via the LED state
  -- vector.
  cmp_bicolor_led_ctrl: bicolor_led_ctrl
    generic map
    (
      g_NB_COLUMN    => 6,
      g_NB_LINE      => 2,
      g_clk_freq     => 125000000,
      g_refresh_rate => 250
    )
    port map
    (
      clk_i           => clk125,
      rst_n_i         => rst_n,
      led_intensity_i => "0011111",
      led_state_i     => bicolor_led_state,
      column_o(0)     => led_wr_ownaddr_i2c_o,
      column_o(1)     => led_wr_gmt_ttl_ttln_o,
      column_o(2)     => led_wr_link_syserror_o,
      column_o(3)     => led_wr_ok_syspw_o,
      column_o(4)     => led_multicast_2_0_o,
      column_o(5)     => led_multicast_3_1_o,
      line_o(0)       => led_ctrl0_o,
      line_o(1)       => led_ctrl1_o,
      line_oen_o(0)   => led_ctrl0_oen_o,
      line_oen_o(1)   => led_ctrl1_oen_o
    );

  --============================================================================
  -- Thermometer test
  --============================================================================
  -- The one-wire master component is used to control the on-board DS18B20
  -- thermometer
  cmp_onewire_master: wb_onewire_master
    generic map
    (
      g_interface_mode      => CLASSIC,
      g_address_granularity => WORD,
      g_num_ports           => 1,
      g_ow_btp_normal       => "5.0",
      g_ow_btp_overdrive    => "1.0"
    )
    port map
    (
      clk_sys_i   => clk125,
      rst_n_i     => rst_n,

      wb_cyc_i    => xbar_master_out(c_slv_onewire_mst).cyc,
      wb_sel_i    => xbar_master_out(c_slv_onewire_mst).sel,
      wb_stb_i    => xbar_master_out(c_slv_onewire_mst).stb,
      wb_we_i     => xbar_master_out(c_slv_onewire_mst).we,
      wb_adr_i    => xbar_master_out(c_slv_onewire_mst).adr(4 downto 2), 
      wb_dat_i    => xbar_master_out(c_slv_onewire_mst).dat,
      wb_dat_o    => xbar_master_in(c_slv_onewire_mst).dat,
      wb_ack_o    => xbar_master_in(c_slv_onewire_mst).ack, 
      wb_int_o    => open, 
      wb_stall_o  => xbar_master_in(c_slv_onewire_mst).stall,

      owr_pwren_o => owr_pwren,
      owr_en_o    => owr_en,
      owr_i       => owr_in
    );
     
  -- Generate tri-state buffer for thermometer
  thermometer_b <= '0' when (owr_en(0) = '1') else 
                   'Z';
  owr_in(0)     <= thermometer_b;

  --============================================================================
  -- Output enable logic
  --============================================================================
  -- The general output enable is set first and the blocking, TTL
  -- and INV output enable signals are set one clock cycle later.
  p_oe: process(clk125)
  begin
    if rising_edge(clk125) then
      if (rst_n = '0') then
        oe     <= '0';
        blo_oe <= '0';
        ttl_oe <= '0';
        inv_oe <= '0';
      else
        oe <= '0';
        if (oe = '1') then
          blo_oe <= '1';
          ttl_oe <= '1';
          inv_oe <= '1';
        end if;
      end if;
    end if;
  end process p_oe;

  fpga_oe_o          <= oe;
  fpga_blo_oe_o      <= blo_oe;
  fpga_trig_ttl_oe_o <= ttl_oe;
  fpga_inv_oe_o      <= inv_oe;
  
  --============================================================================
  -- RTM detection logic
  --============================================================================
  rtmm <= not fpga_rtmm_n_i;
  rtmp <= not fpga_rtmp_n_i;

  cmp_rtm_detector: rtm_detector
    port map
    (
      rtmm_i    => rtmm,
      rtmp_i    => rtmp,
      rtmm_ok_o => rtmm_ok,
      rtmp_ok_o => rtmp_ok
    );

end behav;
